/*
** Der umstlpbare Wrfel nach Paul Schatz
** Copyright (C) 1999,2004 Franz Zahaurek <http://www.fzk.at>
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
*/
/* $Id: wuerfel.h,v 1.13 2004/11/18 23:18:37 fzk Exp $ */

#ifdef CHECKOGLERR
#define wErr( cmd ) \
    cmd; \
    { GLenum error; \
      while ( (error = glGetError()) != GL_NO_ERROR) { \
          fprintf (stderr, "[%s:%d]  %s  failed with error: %s\n", \
                   __FILE__, __LINE__, #cmd, gluErrorString(error) ); \
      } \
    }
#else
#define wErr(cmd) cmd;
#endif

#ifdef OS_WIN32
  #include <windows.h>
#endif
#include <GL/glut.h>
#include <GL/gl.h>

#define FENSTER 600
#define DELTA 0.0001

#define FILE2D(i1, i2) "daten2d"#i1#i2".dat"
#define FILE3D         "daten3d.dat"
#define FILESEGS       "datensegs.dat"
#define FILEOLOID      "datenoloid.dat"

#define ANZPUNKTE 22
#define MAXTFPL 10  /* maximale Anzahl von Punkten fuer eine individuelle Zeitform */

#define MAXOLOID 1000

typedef double CKOORD;            /* eine kartesische Koordinate */
typedef double BWINKEL;           /* ein Winkel im Bogenmass */
#define GWINKEL(b) (double)((b)*180./M_PI)    /* Bogenmass in Grad umrechnen */
#define TOBWINKEL(g) (double)((g)*M_PI/180.)  /* Grad in Bogenmass umrechnen */

#define P(n,i) (wgp->e[n-1].i)    /* Punkt als Pointer */
#define S(n,i) wgp.e[n-1].i       /* Punkt als Struktur */
#define O(n,i) (olo.e[n-1].i)     /* Punkt des Oloids */
#define Oz(n,i) (oloz.e[n-1].i)   /* Punkt des Oloids */

#define KP(n,i) (P(n,i))          /* Koordinate i der Ecke n (Pointer auf Punkt n) */
#define KS(n,i) (S(n,i))          /* Koordinate i der Ecke n (Struktur des Punktes n) */

typedef struct segs {
    char von;                     /* Kante fuehrt vom Punkt von ... */
    char nach;                    /* ... zum Punkt nach             */
} KANTE;

#ifndef OS_WIN32
typedef enum boolean {
    FALSE = 0,
    TRUE = 1,
    nein = FALSE,
    ja = TRUE
} BOOLEAN;
#endif
#define RESET   TRUE
#define ADVANCE FALSE


typedef struct epunkt {      /* ein Punkt in der Ebene */
    CKOORD x;
    CKOORD y;
} EPUNKT;

typedef struct rpunkt {      /* ein Punkt im Raum */
    CKOORD x;
    CKOORD y;
    CKOORD z;
} RPUNKT;

typedef struct wgpos {       /* die Ecken an der Position w */
    BWINKEL w;               /* Winkel des Scharniers P1_P2 mit der x-Achse */
    BWINKEL w2;              /* Winkel der Kante P1_P3 mit der y-Achse
			     ** (= um den das Element 1 von der
			     ** y-Achse weggekippt ist) */
    RPUNKT e[ANZPUNKTE];
} WGPOS;

typedef struct wgplist {     /* der Wuerfelguertel als Liste der Positionen */
    WGPOS wgp;
    struct wgplist *nextwgp;
    struct wgplist *prevwgp;
} WG;

typedef struct oloid {       /* die Enden A und B der Erzeugenden des Oloids */
    RPUNKT e[MAXOLOID];
} OLOID;

typedef struct options {     /* Aufrufparameter */
    unsigned int seite;      /* -s Breite des Fensters in Pixeln */
    BWINKEL      dw;         /* -w Winkel, um den sich 2 Pos. unterscheiden */
    BOOLEAN      print;      /* -p print, gibt Koordinaten der Eckpunkte aus */
    BOOLEAN      virtuell;   /* -e Virtuelle Ecken Evolutions- und Involutionspunkt zeichnen */
    BOOLEAN      ach;        /* -a Hauptachsen x, y, z zeichnen */
    int          mode;       /* -# 1...Hauptdiagonale fixiert
                                   2...Scharnier fixiert
                                   3...Element fixiert
                                   4...Mittelpunkt fixiert */
    BOOLEAN      lines;      /* -g Gitter- oder Flaechendarstellung */
    BOOLEAN      info;       /*    Informations-Subwindow anzeigen (Default) */
    BOOLEAN      tork;       /* -t Wuerfel schwebt */
    BOOLEAN      perspektive;/* -p perspektivische/parallele Projektion */
    BOOLEAN      labels;     /* -l Labels der Ecken zeigen */
    BOOLEAN      kuboid;     /* -k Kuboid zeigen */
    BOOLEAN      oloid;      /* -o Oloid zeigen */
    BOOLEAN      indil;      /* -L von1-bis1,von2-bis2,... indiv. Zeitform Linien */
    BOOLEAN      indip;      /* -P punkt1,punkt21,... indiv. Zeitform der Punkte */
    BOOLEAN      angle1;     /* -W Winkel x-Achse/Scharnier 1-2 anzeigen */
    BOOLEAN      angle2;     /* -H Winkel y-Achse/Kante 1-3 anzeigen */
    BOOLEAN      wuerfel;    /* -w Wuerfelguertel zeigen */
    BOOLEAN      touchline;  /* -b Beruehrungslinie Wuerfelguertel/Timeform zeigen */
    int          position;   /* -P bei dieser Startposition anhalten, -1=nicht gesetzt */
    BOOLEAN      erzeugende; /* -r Erzeugende von Oloid/Kuboid zeigen */
    BOOLEAN      endpunkte;  /* -e Endpunkte von Oloid/Kuboid zeigen */
    int          axistrans;  /* -c # Achsentransf. vor der Darstellung */
                             /*    1...keine, Normalposition = Default*/
                             /*    2...Winkel w um y-Achse */
                             /*    3...Winkel w um y-Achse, dann w2 um x-Achse */
    BOOLEAN      beide;      /* -B Oloid+Kuboid-Berhrungslinie zeigen */
    BOOLEAN      impulse;    /* Leertaste setzt Impuls, schaltet Automatik weiter */
    BOOLEAN      automatic;  /* -A Automatik-Modus */
    BOOLEAN      diagonals;  /* -D Diagonalen des Wuerfels */
    BOOLEAN      center;     /* -M Mittelpunkt des Wuerfels incl. Zeiform */
    BOOLEAN      fullform;   /* -V Oloid/Kuboid voll zeigen (Default=halbe Form) */
} OPTIONS;

typedef struct autooptions { /* Options fuer automatischen Ablauf */
    BOOLEAN      virtuell;   /* Virtuelle Ecken Evolutions- und Involutionspunkt */
    BOOLEAN      ach;        /* Hauptachsen x, y, z */
    int          mode;       /* View-Modus */
    BOOLEAN      lines;      /* Gitter- oder Flaechendarstellung */
    BOOLEAN      info;       /* Informations-Subwindow */
    BOOLEAN      tork;       /* Wuerfel schwebt */
    BOOLEAN      perspektive;/* perspektivische/parallele Projektion */
    BOOLEAN      labels;     /* Labels der Ecken */
    BOOLEAN      kuboid;     /* Kuboid zeigen */
    BOOLEAN      oloid;      /* Oloid zeigen */
    BOOLEAN      wuerfel;    /* Wuerfelguertel */
    BOOLEAN      touchline;  /* Beruehrungslinie Wuerfelguertel/Timeform */
    BOOLEAN      erzeugende; /* Erzeugende von Oloid/Kuboid zeigen */
    BOOLEAN      endpunkte;  /* Endpunkte von Oloid/Kuboid zeigen */
    int          axistrans;  /* Achsentransf. vor der Darstellung */
    BOOLEAN      beide;      /* Oloid+Kuboid-Berhrungslinie zeigen */
    BOOLEAN      diagonals;  /* Diagonalen des Wuerfels */
    BOOLEAN      center;     /* Mittelpunkt des Wuerfels incl. Zeiform */
    char        *msg;        /* Hinweis, was diese Position darstellt */
} AUTOOPTIONS;

#ifdef OS_WIN32
typedef struct timeval {
	int tv_sec;
	int tv_usec;
} TIMEVAL;
#endif

void wPostRedisplay (void);
void wDisplay     (void);
void wReshape     (int w, int h);
void wIdle        (void);
void wKey         (unsigned char key, int x, int y);
void wSpecialKey  (int key, int x, int y);

void wSubDisplay  (void);
void wSubReshape  (int w, int h);

void wDrawString  (char *);
void wCreateDisplayLists (void);
void wTestGlError (char *);

void wCreateMenus (void);
void wMainMenu    (int);
void wSubMenuView (int);
void wSubMenuSpeed (int);
void wSubMenuPoints (int);
void wSubMenuLines (int);
void wExit        (void);
void wToggleTorkeln (void);
void wToggleVirtuell (void);
void wToggleAxis  (void);
void wTogglePerspektive (void);
void wToggleGitterFlaechen (void);
void wToggleView (void);
void wAdvanceAxisTransform (void);
void wAdvanceOptions (BOOLEAN);
void wToggleLabels (void);
void wToggleKuboid (void);
void wToggleOloid (void);
void wToggleBeide (void);
void wToggleWuerfelguertel (void);
void wToggleTouchline (void);
void wToggleErzeugende (void);
void wToggleEndpunkte (void);
void wToggleInfo (void);
void wToggleHalt (void);
void wDisplayOptions (void);
void wResetRotation (void);
void wResetAll (void);
void wSetImpulse (void);
void wToggleAutomatic (void);
void wToggleDiagonals (void);
void wToggleCenter (void);
void wNextPos     (void);
void wPrevPos     (void);
void wThisPos     (void);
void wToggleIndilTimeform (void);
void wToggleIndipTimeform (void);
void wGetTimeformPointsl (void);
void wGetTimeformPointsp (void);
void wToggleAngle1 (void);
void wToggleAngle2 (void);
void wToggleFullform (void);

void wBerechnewg    (void);
void wBerechnewgpos (BWINKEL winkel, WGPOS *wgp);
void wRotatex       (BWINKEL winkel, WGPOS *orig, WGPOS *tran);
void wRotatey       (BWINKEL winkel, WGPOS *orig, WGPOS *tran);
void wRotatez       (BWINKEL winkel, WGPOS *orig, WGPOS *tran);
void wTranslatex    (BWINKEL winkel, WGPOS *orig, WGPOS *tran);
void wTranslatey    (BWINKEL winkel, WGPOS *orig, WGPOS *tran);
void wTranslatez    (BWINKEL winkel, WGPOS *orig, WGPOS *tran);
void wZeichnewg     (void);
void printwg        (void);
void wFreewg        (void);
void wGetOptions    (int argc, char **argv);
void wIdent         (void);
void wUsage         (void);
void wControl       (void);
void wWarten        (int sec);
void torkle         (void);
void wSetOptionDefaults (void);
#ifdef OS_WIN32
double ranf (double init);
#endif

#ifdef DEFINECONSTANTS

WG *wg;            /* erste Position des Wuerfelguertels */
WG *owgp;          /* aktuelle Position */
WG *last;          /* letzte Position */
int anzpos;        /* Anzahl der Positionen des Wrfelgrtels, haengt von dw
		   ** (= Delta des Drehwinkels) ab */

WG *wgw1;          /* analog wg jedoch transferiert um den Winkel w */
WG *owgpw1;        /* aktuelles Element */
WG *lastw1;        /* letztes Element */

WG *wgw2;          /* analog wg jedoch transferiert um die Winkel w + w2 */
WG *owgpw2;        /* aktuelles Element */
WG *lastw2;        /* letztes Element */

/* Zeitform von Linien, ergibt Flaechen */
int tfpl[MAXTFPL]; /* TimeFormPointList: Eckpunkte der Linien der indiv. Zeitform */
int anztfpl;       /* Anzahl der Punkte in der indiv. Zeitform */

/* Zeitform von Punkten, ergibt Linien */
int tfpp[MAXTFPL]; /* TimeFormPoints: Punkte der individuellen Zeitform */
int anztfpp;       /* Anzahl der Punkte in der indiv. Zeitform */

OPTIONS options;   /* Programmoptionen */

GLenum polygonMode;/* Gitter- oder rumliches Modell */

/*
** Die Kanten des Wrfelgrtels liegen jeweils zwischenden zwei Ecken
*/
KANTE kante[] = {
    /* Teil 1 */ { 1, 2}, { 1, 3}, { 1, 4}, { 2, 3}, { 2, 4}, { 3, 4},
    /* Teil 2 */ { 3, 7}, { 3, 8}, { 4, 7}, { 4, 8},
    /* Teil 3 */ { 5, 6}, { 5, 7}, { 5, 8}, { 6, 7}, { 6, 8}, { 7, 8},
    /* Teil 1'*/ /* {11,12},*/ {11,13}, {11,14}, {12,13}, {12,14}, {13,14},
    /* Teil 2'*/ {13,17}, {13,18}, {14,17}, {14,18},
    /* Teil 3'*/ /* {15,16},*/ {15,17}, {15,18}, {16,17}, {16,18}, {17,18}
};
int anzkanten = sizeof (kante) / sizeof (kante[0]);
int anzduenn; /* Anzahl der dnn zu zeichnenden Segmente */

/*
** Oloid aus dem Wuerfelprogramm zeichnen, ein Versuch
*/
/* KANTE koloid[] = { */
/*     { 7, 3}, {3,4}, { 3,17}, {17,18}, {17, 5} */
/* }; */
/* int anzkoloid = sizeof (koloid) / sizeof (koloid[0]); */

/*
** Oloids"Kanten"
       Index     Bedeutung
       1         Ursprung des Koordinatensystems
       2,5       Die Mittelpunkte Ma und Mb (konstant x=-1/2 und x=+1/2)
       3,4       A und B
       6         T
       19        Punkt (2,0,0) auf x-Achse
       20        Punkt (0,2,0) auf y-Achse
       21        Punkt (0,0,2) auf z-Achse
*/
KANTE koloid[] = {
    { 2, 3}, {3, 4}, { 4, 5}, { 6, 3}, { 6, 4}
};
int anzkoloid = sizeof (koloid) / sizeof (koloid[0]);

/*
** Oloid
*/
OLOID olo;
int oi;
int oi1, oi2, oi3, oi4, oi5;

BOOLEAN rechnung_nicht_ok = FALSE;

char *modus_txt[] = { ""
		      , "Hauptdiagonale fixiert"
		      , "Scharnier fixiert"
		      , "Element fixiert"
		      , "Mittelpunkt fixiert"
};

char *darstellung_txt[] = { "Flchen", "Gitter" };

int winIdMain;
int winIdSub;
int menIdMain;
int menIdSubMain;
int menIdMainSubView;
int menIdMainSubSpeed;
int menIdMainSubPoints;
int menIdMainSubLines;

float tranx, trany, tranz;
float rotx, roty, rotz;
float dtranx, dtrany, dtranz, drotx, droty, drotz;
struct timeval waittime;
BOOLEAN  halt;
int listIdLines;
int listIdFill;
int listIdLabels;
int listIdVirt;
int listIdAxis;
int listIdTime1;
int listIdTime2;
int listIdTimeIndil1;
int listIdTimeIndil2;
int listIdTimeIndip1;
int listIdTimeAngle1;
int listIdTimeAngle2;
int listIdTouch;
int listIdDiagonals;
int listIdCenter;
int anzpos;
int listIdLinesStart;
int listIdFillStart;
int listIdLabelsStart;
int listIdVirtStart;
int listIdAxisStart;
int listIdTime1Start;
int listIdTime2Start;
int listIdTimeIndil1Start;
int listIdTimeIndil2Start;
int listIdTimeIndip1Start;
int listIdTimeAngle1Start;
int listIdTimeAngle2Start;
int listIdTouchStart;
int listIdDiagonalsStart;
int listIdCenterStart;
int savedaxistrans;         /* bei automatischer Umschaltung, den alten Wert merken */
#else

extern WG *wg;
extern WG *owgp;
extern WG *last;
extern WG *wgw1;
extern WG *owgpw1;
extern WG *lastw1;
extern WG *wgw2;
extern WG *owgpw2;
extern WG *lastw2;
extern int anzpos;
extern int tfpl[MAXTFPL];    /* TimeFormPointList: Endpunkten von Linien
			     ** des Wuerfelguertels, fuer die die Zeitform
			     ** erstellt werden soll
			     */
extern int anztfpl;
extern int tfpp[MAXTFPL];    /* TimeFormPointList: Punkten
			     ** des Wuerfelguertels, fuer die die Zeitform
			     ** erstellt werden soll
			     */
extern int anztfpp;
extern OPTIONS options;
extern GLenum polygonMode;
extern KANTE kante[];
extern int anzkanten;
extern int anzduenn;
extern KANTE koloid[];
extern int anzkoloid;
extern OLOID olo;
extern int oi;
extern int oi1, oi2, oi3, oi4, oi5;

extern BOOLEAN rechnung_nicht_ok;
extern char* modus_txt[];
extern char* darstellung_txt[];

extern int winIdMain;
extern int winIdSub;
extern int menIdMain;
extern int menIdSubMain;
extern int menIdMainSubView;
extern int menIdMainSubSpeed;
extern int menIdMainSubPoints;
extern int menIdMainSubLines;

extern float tranx, trany, tranz;
extern float rotx, roty, rotz;
extern float dtranx, dtrany, dtranz, drotx, droty, drotz;
extern struct timeval waittime;
extern BOOLEAN  halt;
extern int listIdLines;
extern int listIdFill;
extern int listIdLabels;
extern int listIdVirt;
extern int listIdAxis;
extern int listIdTime1;
extern int listIdTime2;
extern int listIdTimeIndil1;
extern int listIdTimeIndil2;
extern int listIdTimeIndip1;
extern int listIdTimeAngle1;
extern int listIdTimeAngle2;
extern int listIdTouch;
extern int listIdDiagonals;
extern int listIdCenter;
extern int anzpos;
extern int listIdLinesStart;
extern int listIdFillStart;
extern int listIdLabelsStart;
extern int listIdVirtStart;
extern int listIdAxisStart;
extern int listIdTime1Start;
extern int listIdTime2Start;
extern int listIdTimeIndil1Start;
extern int listIdTimeIndil2Start;
extern int listIdTimeIndip1Start;
extern int listIdTimeAngle1Start;
extern int listIdTimeAngle2Start;
extern int listIdTouchStart;
extern int listIdDiagonalsStart;
extern int listIdCenterStart;
extern int savedaxistrans;
#endif
